// This is a sample F# app created in Visual Studio 2012, targeting .NET 4.5

// At time of submission it won't build by default because MD doesn't support .NET 4.5 as yet

[<EntryPoint>]
let main argv = 
    printfn "%A" argv
    0 // return an integer exit code
