
#r "provider_EVIL_PROVIDER_GetInvokerExpression_Null.dll"

open FSharp.EvilProviderWhereGetInvokerExpressionReturnsNull

// It is enough to name the type to expose the validation check
let x = FSharp.EvilProviderWhereGetInvokerExpressionReturnsNull.TheType.Foo
