// Learn more about F# at http://fsharp.org. See the 'F# Tutorial' project
// for more guidance on F# programming.

#load "Library1.fs"
open Sample_VS2015_FSharp_Library_net45

// Define your library scripting code here

