<Application xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
             xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml" 
             x:Class="FSharp.Core.UnitTests.TestApp.App"
             >
    <Application.Resources>
        
    </Application.Resources>
</Application>
