#pragma warning disable 1591
// ------------------------------------------------------------------------------
//  <autogenerated>
//      This code was generated by a tool.
//      Mono Runtime Version: 4.0.30319.17020
// 
//      Changes to this file may cause incorrect behavior and will be lost if 
//      the code is regenerated.
//  </autogenerated>
// ------------------------------------------------------------------------------

[assembly: Android.Runtime.ResourceDesignerAttribute("AndroidWrapper.Resource", IsApplication=true)]

namespace AndroidWrapper
{
	
	
	[System.CodeDom.Compiler.GeneratedCodeAttribute("Xamarin.Android.Build.Tasks", "1.0.0.0")]
	public partial class Resource
	{
		
		static Resource()
		{
			global::Android.Runtime.ResourceIdManager.UpdateIdValues();
		}
		
		public static void UpdateIdValues()
		{
			global::Xamarin.Android.NUnitLite.Resource.Id.OptionHostName = global::AndroidWrapper.Resource.Id.OptionHostName;
			global::Xamarin.Android.NUnitLite.Resource.Id.OptionPort = global::AndroidWrapper.Resource.Id.OptionPort;
			global::Xamarin.Android.NUnitLite.Resource.Id.OptionRemoteServer = global::AndroidWrapper.Resource.Id.OptionRemoteServer;
			global::Xamarin.Android.NUnitLite.Resource.Id.OptionsButton = global::AndroidWrapper.Resource.Id.OptionsButton;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultFullName = global::AndroidWrapper.Resource.Id.ResultFullName;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultMessage = global::AndroidWrapper.Resource.Id.ResultMessage;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultResultState = global::AndroidWrapper.Resource.Id.ResultResultState;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultStackTrace = global::AndroidWrapper.Resource.Id.ResultStackTrace;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultsFailed = global::AndroidWrapper.Resource.Id.ResultsFailed;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultsId = global::AndroidWrapper.Resource.Id.ResultsId;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultsIgnored = global::AndroidWrapper.Resource.Id.ResultsIgnored;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultsInconclusive = global::AndroidWrapper.Resource.Id.ResultsInconclusive;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultsMessage = global::AndroidWrapper.Resource.Id.ResultsMessage;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultsPassed = global::AndroidWrapper.Resource.Id.ResultsPassed;
			global::Xamarin.Android.NUnitLite.Resource.Id.ResultsResult = global::AndroidWrapper.Resource.Id.ResultsResult;
			global::Xamarin.Android.NUnitLite.Resource.Id.RunTestsButton = global::AndroidWrapper.Resource.Id.RunTestsButton;
			global::Xamarin.Android.NUnitLite.Resource.Id.TestSuiteListView = global::AndroidWrapper.Resource.Id.TestSuiteListView;
			global::Xamarin.Android.NUnitLite.Resource.Layout.options = global::AndroidWrapper.Resource.Layout.options;
			global::Xamarin.Android.NUnitLite.Resource.Layout.results = global::AndroidWrapper.Resource.Layout.results;
			global::Xamarin.Android.NUnitLite.Resource.Layout.test_result = global::AndroidWrapper.Resource.Layout.test_result;
			global::Xamarin.Android.NUnitLite.Resource.Layout.test_suite = global::AndroidWrapper.Resource.Layout.test_suite;
		}
		
		public partial class Attribute
		{
			
			static Attribute()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Attribute()
			{
			}
		}
		
		public partial class Drawable
		{
			
			// aapt resource value: 0x7f020000
			public const int Icon = 2130837504;
			
			static Drawable()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Drawable()
			{
			}
		}
		
		public partial class Id
		{
			
			// aapt resource value: 0x7f040001
			public const int OptionHostName = 2130968577;
			
			// aapt resource value: 0x7f040002
			public const int OptionPort = 2130968578;
			
			// aapt resource value: 0x7f040000
			public const int OptionRemoteServer = 2130968576;
			
			// aapt resource value: 0x7f04000f
			public const int OptionsButton = 2130968591;
			
			// aapt resource value: 0x7f04000a
			public const int ResultFullName = 2130968586;
			
			// aapt resource value: 0x7f04000c
			public const int ResultMessage = 2130968588;
			
			// aapt resource value: 0x7f04000b
			public const int ResultResultState = 2130968587;
			
			// aapt resource value: 0x7f04000d
			public const int ResultStackTrace = 2130968589;
			
			// aapt resource value: 0x7f040006
			public const int ResultsFailed = 2130968582;
			
			// aapt resource value: 0x7f040003
			public const int ResultsId = 2130968579;
			
			// aapt resource value: 0x7f040007
			public const int ResultsIgnored = 2130968583;
			
			// aapt resource value: 0x7f040008
			public const int ResultsInconclusive = 2130968584;
			
			// aapt resource value: 0x7f040009
			public const int ResultsMessage = 2130968585;
			
			// aapt resource value: 0x7f040005
			public const int ResultsPassed = 2130968581;
			
			// aapt resource value: 0x7f040004
			public const int ResultsResult = 2130968580;
			
			// aapt resource value: 0x7f04000e
			public const int RunTestsButton = 2130968590;
			
			// aapt resource value: 0x7f040010
			public const int TestSuiteListView = 2130968592;
			
			static Id()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Id()
			{
			}
		}
		
		public partial class Layout
		{
			
			// aapt resource value: 0x7f030000
			public const int options = 2130903040;
			
			// aapt resource value: 0x7f030001
			public const int results = 2130903041;
			
			// aapt resource value: 0x7f030002
			public const int test_result = 2130903042;
			
			// aapt resource value: 0x7f030003
			public const int test_suite = 2130903043;
			
			static Layout()
			{
				global::Android.Runtime.ResourceIdManager.UpdateIdValues();
			}
			
			private Layout()
			{
			}
		}
	}
}
#pragma warning restore 1591
