module A

let X1 (x:int) = x + 1
let X2 (x:int) = x + 1
let X3 (x:int) = x + 1
let X4 (x:int) = x + 1)
let X5 (x:int) = x + 1
let X6 (x:int) = x + 1
let X7 (x:int) = x + 1
let X8 (x:int) = x + 1
let X9 (x:int) = x + 1
let X10 (x:int) = x + 1

let Y1 (x:int) = x + 1
let Y2 (x:int) = x + 1
let Y3 (x:int) = x + 1
let Y4 (x:int) = x + 1
let Y5 (x:int) = x + 1
let Y6 (x:int) = x + 1
let Y7 (x:int) = x + 1
let Y8 (x:int) = x + 1
let Y9 (x:int) = x + 1
let Y10 (x:int) = x + 1


