#r "provider_EVIL_PROVIDER_GetStaticParameters_Exception.dll"


open FSharp.EvilProviderWhereGetStaticParametersRaisesException

// It is enough to name the type to expose the validation check
type Negative1 = FSharp.EvilProviderWhereGetStaticParametersRaisesException.TheType
