module Neg53

module Test1 = 
    type D() = 

       static let x = 1

 

    D.``.cctor``()


module Test2 = 
    type D() = 

       static let x = 1


    D.``.ctor``()

module Test3 = 
    type D() = 

       static member ``.ctor``() = 1

