// Copyright (c) Microsoft Open Technologies, Inc.  All Rights Reserved.  Licensed under the Apache License, Version 2.0.  See License.txt in the project root for license information.

module Microsoft.FSharp.Compiler.Interactive.Attributes
[<assembly: AutoOpen("Microsoft.FSharp.Compiler.Interactive.Settings")>]
do()

