// Learn more about F# at http://fsharp.net. See the 'F# Tutorial' project
// for more guidance on F# programming.

#load "PortableLibrary1.fs"
open Sample_VS2013_FSharp_Portable_Library_Legacy_net45

