// run as "fsi --exec CreateComparersRegression.fsx > tmp.txt"
// and then copy text in tmp.txt into the bottom of ComparersRegression.fs

#r @"..\..\..\..\Release\net40\bin\FSharp.Core.Unittests.dll"

FSharp.Core.Unittests.FSharp_Core.Microsoft_FSharp_Core.ComparersRegression.createData ()