#r "provider_EVIL_PROVIDER_GetStaticParameters_Null.dll"


open FSharp.EvilProviderWhereGetStaticParametersReturnsNull

// It is enough to name the type to expose the validation check
type Negative1 = FSharp.EvilProviderWhereGetStaticParametersReturnsNull.TheType
