
//------------------------------------------------------------------------------
// <autogenerated>
//     This code was generated by a tool.
//     Runtime Version: 4.0.30319.17020
//
//     Changes to this file may cause incorrect behavior and will be lost if 
//     the code is regenerated.
// </autogenerated>
//------------------------------------------------------------------------------

namespace global
[<assembly: Android.Runtime.ResourceDesignerAttribute("Dummy6.Resource", IsApplication=true)>]
    do ()

namespace AndroidFSharpLanguageUnitTests
    // Generated by F# CodeDom
    #nowarn "49" // uppercase argument names
    #nowarn "67" // this type test or downcast will always hold
    #nowarn "66" // this upcast is unnecessary - the types are identical
    #nowarn "58" // possible incorrect indentation..
    #nowarn "57" // do not use create_DelegateEvent
    #nowarn "51" // address-of operator can occur in the code
    #nowarn "1183" // unused 'this' reference
    
    exception ReturnException4ba235d12fd842d990687877f398e5e9 of obj
    exception ReturnNoneException4ba235d12fd842d990687877f398e5e9
    [<AutoOpen>]
    module FuncConvertFinalOverload4ba235d12fd842d990687877f398e5e9 =
      // This extension member adds to the FuncConvert type and is the last resort member in the method overloading rules. 
      type global.Microsoft.FSharp.Core.FuncConvert with
          /// A utility function to convert function values from tupled to curried form
          static member FuncFromTupled (f:'T -> 'Res) = f
    
    type
        [<System.CodeDom.Compiler.GeneratedCodeAttribute("Xamarin.Android.Build.Tasks", "1.0.0.0")>]
        
        (* partial *)Resource () =
            static do Android.Runtime.ResourceIdManager.UpdateIdValues()
            (* Member of type 'CodeTypeConstructor' is not supported by the CodeDOM provider and was omitted *)
            static member UpdateIdValues  () =
                global.Xamarin.Android.NUnitLite.Resource.Id.OptionHostName <- Resource_Id.OptionHostName
                global.Xamarin.Android.NUnitLite.Resource.Id.OptionPort <- Resource_Id.OptionPort
                global.Xamarin.Android.NUnitLite.Resource.Id.OptionRemoteServer <- Resource_Id.OptionRemoteServer
                global.Xamarin.Android.NUnitLite.Resource.Id.OptionsButton <- Resource_Id.OptionsButton
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultFullName <- Resource_Id.ResultFullName
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultMessage <- Resource_Id.ResultMessage
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultResultState <- Resource_Id.ResultResultState
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultStackTrace <- Resource_Id.ResultStackTrace
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultsFailed <- Resource_Id.ResultsFailed
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultsId <- Resource_Id.ResultsId
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultsIgnored <- Resource_Id.ResultsIgnored
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultsInconclusive <- Resource_Id.ResultsInconclusive
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultsMessage <- Resource_Id.ResultsMessage
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultsPassed <- Resource_Id.ResultsPassed
                global.Xamarin.Android.NUnitLite.Resource.Id.ResultsResult <- Resource_Id.ResultsResult
                global.Xamarin.Android.NUnitLite.Resource.Id.RunTestsButton <- Resource_Id.RunTestsButton
                global.Xamarin.Android.NUnitLite.Resource.Id.TestSuiteListView <- Resource_Id.TestSuiteListView
                global.Xamarin.Android.NUnitLite.Resource.Layout.options <- Resource_Layout.options
                global.Xamarin.Android.NUnitLite.Resource.Layout.results <- Resource_Layout.results
                global.Xamarin.Android.NUnitLite.Resource.Layout.test_result <- Resource_Layout.test_result
                global.Xamarin.Android.NUnitLite.Resource.Layout.test_suite <- Resource_Layout.test_suite


    
    and
        
        (* partial *)Resource_Attribute () =
            static do Android.Runtime.ResourceIdManager.UpdateIdValues()
            (* Member of type 'CodeTypeConstructor' is not supported by the CodeDOM provider and was omitted *)


    
    and
        
        (* partial *)Resource_Drawable () =
            static do Android.Runtime.ResourceIdManager.UpdateIdValues()
            // aapt resource value: 0x7f020000
            static member Icon = 2130837504
            (* Member of type 'CodeTypeConstructor' is not supported by the CodeDOM provider and was omitted *)


    
    and
        
        (* partial *)Resource_Id () =
            static do Android.Runtime.ResourceIdManager.UpdateIdValues()
            // aapt resource value: 0x7f040001
            static member OptionHostName = 2130968577
            
            // aapt resource value: 0x7f040002
            static member OptionPort = 2130968578
            
            // aapt resource value: 0x7f040000
            static member OptionRemoteServer = 2130968576
            
            // aapt resource value: 0x7f04000f
            static member OptionsButton = 2130968591
            
            // aapt resource value: 0x7f04000a
            static member ResultFullName = 2130968586
            
            // aapt resource value: 0x7f04000c
            static member ResultMessage = 2130968588
            
            // aapt resource value: 0x7f04000b
            static member ResultResultState = 2130968587
            
            // aapt resource value: 0x7f04000d
            static member ResultStackTrace = 2130968589
            
            // aapt resource value: 0x7f040006
            static member ResultsFailed = 2130968582
            
            // aapt resource value: 0x7f040003
            static member ResultsId = 2130968579
            
            // aapt resource value: 0x7f040007
            static member ResultsIgnored = 2130968583
            
            // aapt resource value: 0x7f040008
            static member ResultsInconclusive = 2130968584
            
            // aapt resource value: 0x7f040009
            static member ResultsMessage = 2130968585
            
            // aapt resource value: 0x7f040005
            static member ResultsPassed = 2130968581
            
            // aapt resource value: 0x7f040004
            static member ResultsResult = 2130968580
            
            // aapt resource value: 0x7f04000e
            static member RunTestsButton = 2130968590
            
            // aapt resource value: 0x7f040010
            static member TestSuiteListView = 2130968592
            (* Member of type 'CodeTypeConstructor' is not supported by the CodeDOM provider and was omitted *)


    
    and
        
        (* partial *)Resource_Layout () =
            static do Android.Runtime.ResourceIdManager.UpdateIdValues()
            // aapt resource value: 0x7f030000
            static member options = 2130903040
            
            // aapt resource value: 0x7f030001
            static member results = 2130903041
            
            // aapt resource value: 0x7f030002
            static member test_result = 2130903042
            
            // aapt resource value: 0x7f030003
            static member test_suite = 2130903043
            (* Member of type 'CodeTypeConstructor' is not supported by the CodeDOM provider and was omitted *)

